//                                               -*- C++ -*-
/**
 * @file  ComputedNumericalMathEvaluationImplementation.cxx
 * @brief Abstract top-level class for all distributions
 *
 * (C) Copyright 2005-2006 EDF
 *
 * Permission to copy, use, modify, sell and distribute this software
 * is granted provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied
 * warranty, and with no claim as to its suitability for any purpose.
 *
 *
 * \author $LastChangedBy: dutka $
 * \date   $LastChangedDate: 2009-05-28 14:47:53 +0200 (jeu. 28 mai 2009) $
 */

#include "ComputedNumericalMathEvaluationImplementation.hxx"
#include "ComputedNumericalMathEvaluationImplementationFactory.hxx"
#include "PersistentObjectFactory.hxx"
#include "Cache.hxx"
#include "Log.hxx"
#include "WrapperData.hxx"

namespace OpenTURNS {

  namespace Base {

    namespace Func {

      using Common::Log;

      CLASSNAMEINIT(ComputedNumericalMathEvaluationImplementation);

      static Common::Factory<ComputedNumericalMathEvaluationImplementation> RegisteredFactory("ComputedNumericalMathEvaluationImplementation");

      /* Default constructor */
      ComputedNumericalMathEvaluationImplementation::ComputedNumericalMathEvaluationImplementation(const String & name,
												   const WrapperFile & file)
	throw(WrapperInternalException)	:
	NumericalMathEvaluationImplementation(),
	p_function_(0),
	p_state_(0)
      {
	setName(name);
	const WrapperData data = file.getWrapperData();
	if (! data.isValid()) throw WrapperInternalException(HERE) << "The wrapper data are not valid";

	if (data.getFunctionDescription().provided_) {
	  p_function_.reset(new WrapperObject( data.getLibraryPath(),
					       data.getFunctionDescription().name_,
					       data,
					       WrapperObject::FUNCTION ));
	}

	if (p_function_.isNull()) throw WrapperInternalException(HERE) << "Unable to allocate wrapper";

	// Initialize the state into the wrapper
	p_state_ = p_function_->createNewState();
	NumericalMathEvaluationImplementation::setDescription(getDescription());

	// Activate the cache because external codes are supposed expensive
	NumericalMathEvaluationImplementation::enableCache();
      }


      /* Copy constructor */
      ComputedNumericalMathEvaluationImplementation::ComputedNumericalMathEvaluationImplementation(const ComputedNumericalMathEvaluationImplementation & other)
	throw(WrapperInternalException)
	: NumericalMathEvaluationImplementation(other),
	  p_function_(other.p_function_),
	  p_state_(0)
      {
	if (p_function_.isNull()) throw WrapperInternalException(HERE) << "Unable to allocate wrapper";

	// Initialize the state into the wrapper
	p_state_ = p_function_->createNewState();
      }


      /* Virtual constructor */
      ComputedNumericalMathEvaluationImplementation * ComputedNumericalMathEvaluationImplementation::clone() const
      {
	return new ComputedNumericalMathEvaluationImplementation(*this);
      }

      /* Destructor */
      ComputedNumericalMathEvaluationImplementation::~ComputedNumericalMathEvaluationImplementation()
      {
	// Call the finalization function before destruction
	p_function_->finalize( p_state_ );

	// Delete the state into the wrapper
	p_function_->deleteState( p_state_ );
      }


      /* Comparison operator */
      Bool ComputedNumericalMathEvaluationImplementation::operator ==(const ComputedNumericalMathEvaluationImplementation & other) const
      {
	return true;
      }

  
      /* String converter */
      String ComputedNumericalMathEvaluationImplementation::__repr__() const {
	OSS oss;
	oss << "class=" << ComputedNumericalMathEvaluationImplementation::GetClassName()
	    << " name=" << getName();
	return oss;
      }

  



      /* State accessor */
      void * ComputedNumericalMathEvaluationImplementation::getState() const
      {
	return p_state_;
      }






      /* Here is the interface that all derived class must implement */
	
      /* Operator () */
      ComputedNumericalMathEvaluationImplementation::NumericalPoint
      ComputedNumericalMathEvaluationImplementation::operator() (const NumericalPoint & in) const
	throw(InvalidArgumentException,InternalException)
      {
	CacheKeyType inKey = in.getCollection();
	try {
	  // First, initialize the external code on first invocation
	  if (callsNumber_ == 0) p_function_->initialize( p_state_ );

	  // Execute the wrapper
	  if ( p_cache_->hasKey( inKey ) )
	    return NumericalPoint::ImplementationType( p_cache_->find( inKey ) );
	  else {
	    ++callsNumber_;
	    NumericalPoint out = p_function_->execute( p_state_, in );
	    CacheValueType outValue = out.getCollection();
	    if (p_cache_->isEnabled()) p_cache_->add( inKey, outValue );
	    return out;
	  }
	}
	catch (InvalidArgumentException & ex) {
	  throw;
	}
 	catch (WrapperInternalException & ex) {
	  throw InternalException(HERE) << ex;
	}
      }

      /* Operator () */
      ComputedNumericalMathEvaluationImplementation::NumericalSample
      ComputedNumericalMathEvaluationImplementation::operator() (const NumericalSample & in) const
	throw(InvalidArgumentException, InternalException)
      {

	try {
	  // First, initialize the external code on first invocation
	  if (callsNumber_ == 0) p_function_->initialize( p_state_ );

	  if (p_cache_->isEnabled()) {
	    // We build a new sample without the already computed points
	    NumericalSample newIn( 0, in.getDimension() );
	    for(UnsignedLong i=0; i<in.getSize(); ++i) {
	      if ( ! p_cache_->hasKey( in[i].getCollection() ) ) newIn.add( in[i] );
	    }


	    // Execute the wrapper if needed
	    NumericalSample newOut( 0, getOutputNumericalPointDimension() );
	    if (newIn.getSize() > 0) {

	      // We shrink newIn and remove double entries
	      std::sort( newIn.begin(), newIn.end() );
	      NumericalSample::iterator last = std::unique( newIn.begin(), newIn.end() );
	      newIn.erase( last, newIn.end() );
	    
	      callsNumber_ += newIn.getSize();
	      newOut = p_function_->execute( p_state_, newIn );
	    }

	    // We use a secondary cache to hold the computed values
	    CacheType tempCache( newIn.getSize() );
	    tempCache.enable();
	    for(UnsignedLong i=0; i<newIn.getSize(); ++i) tempCache.add( newIn[i].getCollection(), newOut[i].getCollection() );

	    // We gather the computed values into a new output sample
	    NumericalSample out( in.getSize(), getOutputNumericalPointDimension() );
	    for(UnsignedLong i=0; i<in.getSize(); ++i) {
	      if ( p_cache_->hasKey( in[i].getCollection() ) ) {
		out[i] = NumericalPoint::ImplementationType( p_cache_->find( in[i].getCollection() ) );

	      } else if ( tempCache.hasKey( in[i].getCollection() ) ) {
		out[i] = NumericalPoint::ImplementationType( tempCache.find( in[i].getCollection() ) );

	      } else throw InternalException(HERE) << "Error in computing sample. Some values can not be retrieved from cache.";
	    
	    }

	    // We add the computed values into the cache AFTER having read the cache because
	    // older values may be flushed
	    //for(UnsignedLong i=0; i<newIn.getSize(); ++i) p_cache_->add( newIn[i], newOut[i] );
	    p_cache_->merge( tempCache );

	    // return the gathered sample
	    return out;

	  } else {
	    callsNumber_ += in.getSize();
	    return p_function_->execute( p_state_, in );
	  }

	}
	catch (InvalidArgumentException & ex) {
	  throw;
	}
 	catch (WrapperInternalException & ex) {
	  throw InternalException(HERE) << ex;
	}
      }



      /* Accessor for input point dimension */
      UnsignedLong ComputedNumericalMathEvaluationImplementation::getInputNumericalPointDimension() const
	throw(InternalException)
      {
	UnsignedLong inputDimension = 0;

	try {
	  inputDimension = p_function_->getInNumericalPointDimension( p_state_ );
	}
 	catch (WrapperInternalException & ex) {
	  throw InternalException(HERE) << ex;
	}

	return inputDimension;
      }



      /* Accessor for output point dimension */
      UnsignedLong ComputedNumericalMathEvaluationImplementation::getOutputNumericalPointDimension() const
	throw(InternalException)
      {
	UnsignedLong outputDimension = 0;

	try {
	  outputDimension = p_function_->getOutNumericalPointDimension( p_state_ );
	}
 	catch (WrapperInternalException & ex) {
	  throw InternalException(HERE) << ex;
	}

	return outputDimension;
      }

      /* Accessor for output point dimension */
      ComputedNumericalMathEvaluationImplementation::Description ComputedNumericalMathEvaluationImplementation::getDescription() const
	throw(InternalException)
      {
	Description description;

	try {
	  // Here, we get only the description of the input variable
	  description = p_function_->getDescription();
	  if (description.getSize() == getInputNumericalPointDimension())
	    {
	      // Put generic names for the output description if something they are missing in the wrapper
	      for (UnsignedLong i = 0; i < getOutputNumericalPointDimension(); ++i) description.add(OSS() << "y" << i);
	    }
	  // If the description does not match the dimensions, error
	  if (description.getSize() != getInputNumericalPointDimension() + getOutputNumericalPointDimension()) throw InternalException(HERE) << "Error: the description " << description << " does not match the dimensions of the function. Here, input dimension=" << getInputNumericalPointDimension() << " and output dimension=" << getOutputNumericalPointDimension() << ". Check the wrapper description.";
	}
 	catch (WrapperInternalException & ex) {
	  throw InternalException(HERE) << ex;
	}

	return description;
      }
      
      /* Method save() stores the object through the StorageManager */
      void ComputedNumericalMathEvaluationImplementation::save(const StorageManager::Advocate & adv) const
      {
	NumericalMathEvaluationImplementation::save(adv);
      }

      /* Method load() reloads the object from the StorageManager */
      void ComputedNumericalMathEvaluationImplementation::load(const StorageManager::Advocate & adv)
      {
	NumericalMathEvaluationImplementation::load(adv);
	*this = *static_cast<ComputedNumericalMathEvaluationImplementation*>(ComputedNumericalMathEvaluationImplementationFactory::getInstance()->buildObject(getName()));
      }

    } /* namespace Func */
  } /* namespace Base */
} /* namespace OpenTURNS */
